#include "saadc.h"
#include <nrfx_saadc.h>
#include <nrfx_rtc.h>
#include <nrfx_ppi.h>

/***************************************************************************************/

#define ADC_CHANNELS_IN_USE     2							
#define SAADC_BUF_SIZE          32							/*No of samples per buffer given out by SAADC*/
											
#define SAADC_BUF_COUNT         3							/*Number of buffer given to SAAdC*/
#define SAADC_SAMPLE_FREQUENCY  512							/*Sampling frequency*/

/***************************************************************************************/

#define SAADC_CHANNEL1_INDEX 0
#define SAADC_CHANNEL2_INDEX 1

#define SAADC_CHANNEL1_MASK (1U << SAADC_CHANNEL1_INDEX)
#define SAADC_CHANNEL2_MASK (1U << SAADC_CHANNEL2_INDEX)

uint32_t saadc_sample_task_address;

/***************************************************************************************/

/*Configurations and buffers SAADC, TIMER and PPI*/
static nrf_saadc_value_t samples[SAADC_BUF_COUNT][SAADC_BUF_SIZE];			/*Buffers for SAADC sampling*/

static const nrfx_rtc_t m_sample_rtc = NRFX_RTC_INSTANCE(2);
static nrf_ppi_channel_t m_timer_saadc_ppi_channel;
static nrf_ppi_channel_t m_end_start_ppi_channel;
nrfx_err_t err_code;

/***************************************************************************************/

void print_array(nrf_saadc_value_t *buf, uint16_t len)
{
	printk("saadc op : ");
	for(int i=0;i<len;i++)
	{
		printk("%d, ", buf[i]);
	}
	printk("\n");
}

/***************************************************************************************/

/**
 * @brief For assigning a buffer to the ADC.
 *
 * retval index of the buffer in static nrf_saadc_value_t samples
 */
static void set_buffer(void)
{
    static int buffer_index = -1;
    buffer_index = (buffer_index + 1) % SAADC_BUF_COUNT;

    err_code = nrfx_saadc_buffer_set((samples[buffer_index]), SAADC_BUF_SIZE);
	if(err_code!=NRFX_SUCCESS){
		printk("Error in setting buffer: %02x\n", err_code);
	}
}

/***************************************************************************************/
/*Added a ppi channel to attach end event to start task*/
static void event_handler(nrfx_saadc_evt_t const * p_event)
{
	static uint32_t prv_ts = 0;
	uint32_t curr_ts = k_cycle_get_32();
	printk("\r\nTIME DIFF : %d, SAADC EVENT - %d\r\n", curr_ts-prv_ts, p_event->type);
	prv_ts = curr_ts;

    switch (p_event->type)
    {
        case NRFX_SAADC_EVT_DONE:
		{
			print_array(p_event->data.done.p_buffer, p_event->data.done.size);
			break;
		}

        case NRFX_SAADC_EVT_BUF_REQ:
            // Set up the next available buffer
           set_buffer();
		   printk("buffer set\n");
        break;

		default:
		break;
    }
}

int configure_channel1()
{
	int err_code;
	static nrfx_saadc_channel_t channel1_config;

	channel1_config.channel_config.resistor_p = NRF_SAADC_RESISTOR_DISABLED;      /*Resistor Pull UP/Down Disabled*/
	channel1_config.channel_config.resistor_n = NRF_SAADC_RESISTOR_DISABLED;      /*Resistor Pull Up/Down Disabled*/
	channel1_config.channel_config.reference  = NRF_SAADC_REFERENCE_VDD4;         /*Set VDD/4 as SAADC Reference*/
	channel1_config.channel_config.gain       = NRF_SAADC_GAIN1_4;                /*Set the Gain Factor as 1/4*/
	channel1_config.channel_config.acq_time   = NRF_SAADC_ACQTIME_40US;           /*Set SAADC Acquisition Time as 40us. Source resistance is ~ 800K ohms.*/
	channel1_config.channel_config.mode       = NRF_SAADC_MODE_SINGLE_ENDED;      /*Set he SAADC Acquisitonn Mode as Single Ended*/
	channel1_config.pin_n                     = NRF_SAADC_INPUT_DISABLED;         /*Disable the second SAADC input (Single Ended)*/
	channel1_config.pin_p                     = NRF_SAADC_INPUT_AIN7;             /*Enable the second SAADC input (Single Ended). */
	channel1_config.channel_config.burst      = NRF_SAADC_BURST_DISABLED;         /*Disable Burst Mode*/
	channel1_config.channel_index             = SAADC_CHANNEL1_INDEX;

	err_code = nrfx_saadc_channel_config(&channel1_config);
	if(err_code!=NRFX_SUCCESS)
	{
		printk("Error in saadc channel config: %02x\n", err_code);
		return err_code;
	}
	printk("sampling channel 1 configured\n");
	return 0;
}

/* Function to configure PPI */
 void ppi_init(void)
{
  // Trigger task sample from timer
  nrfx_err_t err_code = nrfx_ppi_channel_alloc(&m_timer_saadc_ppi_channel);
  if(err_code!=NRFX_SUCCESS)
  {
    printk(" PPI channel allocation error %d",err_code);
  }
  err_code = nrfx_ppi_channel_assign(m_timer_saadc_ppi_channel, 
                                     nrfx_rtc_event_address_get(&m_sample_rtc, NRF_RTC_EVENT_TICK),
                                     nrf_saadc_task_address_get((NRF_SAADC_Type *)0x40007000,NRF_SAADC_TASK_SAMPLE));
  if(err_code!=NRFX_SUCCESS)
  {
    printk(" PPI channel assignment error %d",err_code);
  }

  err_code = nrfx_ppi_channel_enable(m_timer_saadc_ppi_channel);

  if(err_code!=NRFX_SUCCESS)
  {
    printk(" PPI channel enable error %d",err_code);
  }

	/***********************************************************************************/
		err_code = nrfx_ppi_channel_alloc(&m_end_start_ppi_channel);
	if(err_code!=NRFX_SUCCESS)
	{
		printk(" PPI channel allocation error %d",err_code);
	}
	err_code = nrfx_ppi_channel_assign(m_end_start_ppi_channel, 
										nrf_saadc_event_address_get((NRF_SAADC_Type *)0x40007000,NRF_SAADC_EVENT_END),
										nrf_saadc_task_address_get((NRF_SAADC_Type *)0x40007000,NRF_SAADC_TASK_START));
	if(err_code!=NRFX_SUCCESS){
		printk(" PPI channel assignment error %d",err_code);
	}

	err_code = nrfx_ppi_channel_enable(m_end_start_ppi_channel);
	if(err_code!=NRFX_SUCCESS){
		printk(" PPI channel enable error %d",err_code);
	}

  /*************************************************************************************/
  printk("ppi done \n");
}

static void rtc_handler(nrfx_rtc_int_type_t int_type){}

void rtc_init(void)
{
	nrfx_err_t err_code;

	nrfx_rtc_config_t rtc_config = NRFX_RTC_DEFAULT_CONFIG;
	rtc_config.prescaler = RTC_FREQ_TO_PRESCALER(SAADC_SAMPLE_FREQUENCY);

	err_code = nrfx_rtc_init(&m_sample_rtc, &rtc_config, rtc_handler);
	if(err_code!=NRFX_SUCCESS)
	{
		printk(" rtc initialization error %d",err_code);
	}

	/*enabling overflow interrupt for ticking 32 counter.
		The RTC is counter 24 bit counter.*/
	nrfx_rtc_overflow_enable(&m_sample_rtc, true);

	IRQ_CONNECT(DT_IRQN(DT_NODELABEL(rtc2)),
				DT_IRQ(DT_NODELABEL(rtc2), priority),
				nrfx_rtc_2_irq_handler, NULL, 0);

	printk("RTC initialized\n");
}

int configure_channel2()
{
	int err_code;
	static nrfx_saadc_channel_t channel2_config;

	channel2_config.channel_config.resistor_p = NRF_SAADC_RESISTOR_DISABLED;      /*Resistor Pull UP/Down Disabled*/
	channel2_config.channel_config.resistor_n = NRF_SAADC_RESISTOR_DISABLED;      /*Resistor Pull Up/Down Disabled*/
	channel2_config.channel_config.reference  = NRF_SAADC_REFERENCE_VDD4;         /*Set VDD/4 as SAADC Reference*/
	channel2_config.channel_config.gain       = NRF_SAADC_GAIN1_4;                /*Set the Gain Factor as 1/4*/
	channel2_config.channel_config.acq_time   = NRF_SAADC_ACQTIME_40US;           /*Set SAADC Acquisition Time as 40us. Source resistance is ~ 800K ohms.*/
	channel2_config.channel_config.mode       = NRF_SAADC_MODE_SINGLE_ENDED;      /*Set he SAADC Acquisitonn Mode as Single Ended*/
	channel2_config.pin_n                     = NRF_SAADC_INPUT_DISABLED;         /*Disable the second SAADC input (Single Ended)*/
	channel2_config.pin_p                     = NRF_SAADC_INPUT_AIN3;             /*Enable the second SAADC input (Single Ended).*/
	channel2_config.channel_config.burst      = NRF_SAADC_BURST_DISABLED;         /*Disable Burst Mode*/
	channel2_config.channel_index             = SAADC_CHANNEL2_INDEX;

	err_code = nrfx_saadc_channel_config(&channel2_config);
	if(err_code!=NRFX_SUCCESS)
	{
		printk("Error in saadc channel config: %02x\n", err_code);
		return err_code;
	}
	printk("sampling channel 2 configured\n");
	return 0;
}

int configure_dual_channel_saadc()
{
	uint32_t channel_mask = SAADC_CHANNEL1_MASK | SAADC_CHANNEL2_MASK;

	nrfx_saadc_adv_config_t saadc_adv_config = NRFX_SAADC_DEFAULT_ADV_CONFIG;
	saadc_adv_config.internal_timer_cc = 0;
	saadc_adv_config.start_on_end = false;

	IRQ_CONNECT(DT_IRQN(DT_NODELABEL(adc)),
				DT_IRQ(DT_NODELABEL(adc), priority),
				nrfx_saadc_irq_handler, NULL, 0);

	err_code = nrfx_saadc_init(NRFX_SAADC_DEFAULT_CONFIG_IRQ_PRIORITY);
	if(err_code !=NRFX_SUCCESS)
	{
		printk("Error in saadc setup: %02x\n", err_code);
	}

	configure_channel1();
	configure_channel2();

	err_code = nrfx_saadc_advanced_mode_set(channel_mask,
											NRF_SAADC_RESOLUTION_12BIT,
											&saadc_adv_config,
											event_handler);
	if(err_code!=NRFX_SUCCESS)
	{
		printk("Error in saadc advanced mode config: %02x\n", err_code);
	}
                                        
  // Configure two buffers to ensure double buffering of samples, to avoid data loss when the sampling frequency is high
  set_buffer();
  set_buffer();

  printk("done saadc");

  ppi_init();

  rtc_init();

  err_code = nrfx_saadc_mode_trigger();
  if(err_code!=NRFX_SUCCESS)
  {
      printk("Error in saadc mode trigger: %02x\n", err_code);
  }

  return 0;
  }

/*************************************************************************************************************/

/** @brief initiates adc sampling.
 */
void start_sampling()
{
	nrfx_rtc_enable(&m_sample_rtc);
	nrfx_rtc_tick_enable(&m_sample_rtc, true);
	printk("sampling started\n");

}

/*************************************************************************************************************/

/** 
 * @brief stops any ongoing streaming to the app.
 */
void stop_sampling()
{
	nrfx_rtc_disable(&m_sample_rtc);
	nrfx_rtc_tick_disable(&m_sample_rtc);
	printk("sampling stopped\n");
}

/*************************************************************************************************************/
